defmodule ClickHouse.Query do
  @moduledoc """
  A ClickHouse query.
  """

  alias ClickHouse.Config

  @type command :: :create | :select | :insert | :alter | :unknown
  @type t :: %__MODULE__{
          client: ClickHouse.client(),
          interface: ClickHouse.interface() | nil,
          format: ClickHouse.format() | nil,
          command: command(),
          statement: ClickHouse.statement(),
          encoded: binary() | nil,
          params: ClickHouse.params(),
          params_count: non_neg_integer()
        }

  @format_regex ~r/(?<= (?:FORMAT|format) )(\w+)/
  @create_regex ~r/\bCREATE\b/i
  @select_regex ~r/\bSELECT\b/i
  @insert_regex ~r/\bINSERT\b/i
  @alter_regex ~r/\bALTER\b/i
  @replace_regex ~r/\$\d+/

  defstruct [
    :client,
    :interface,
    :format,
    :command,
    :statement,
    :encoded,
    :params,
    :params_count,
    :opts
  ]

  ################################
  # Public API
  ################################

  @spec new(
          ClickHouse.client(),
          ClickHouse.statement(),
          ClickHouse.params(),
          keyword()
        ) ::
          ClickHouse.Query.t()
  def new(client, statement, params \\ [], opts \\ []) do
    %__MODULE__{
      client: client,
      statement: statement,
      params: params,
      opts: opts
    }
    |> put_interface()
    |> parse_format()
    |> parse_command()
    |> parse_params_count()
  end

  @spec put_encoded(ClickHouse.Query.t(), binary()) :: ClickHouse.Query.t()
  def put_encoded(query, encoded) do
    %{query | encoded: encoded}
  end

  @spec get_statement(ClickHouse.Query.t()) :: binary()
  def get_statement(query) do
    query.encoded || query.statement
  end

  @spec encode(ClickHouse.Query.t()) ::
          {:ok, ClickHouse.Query.t()} | {:error, ClickHouse.FormatError.t()}
  def encode(%{format: nil} = query), do: {:ok, query}
  def encode(query), do: query.format.encode(query)

  ################################
  # Private API
  ################################

  defp put_interface(query) do
    case Config.interface(query.client) do
      {:ok, interface} -> %{query | interface: interface}
      _ -> query
    end
  end

  defp parse_format(query) do
    case Regex.run(@format_regex, query.statement, capture: :first) do
      [format] -> get_format(query, format)
      _ -> query
    end
  end

  defp get_format(query, format) do
    case Config.format(query.client, format) do
      {:ok, format} -> %{query | format: format}
      _ -> query
    end
  end

  defp parse_command(query) do
    with {_, false} <- {:create, Regex.match?(@create_regex, query.statement)},
         {_, false} <- {:select, Regex.match?(@select_regex, query.statement)},
         {_, false} <- {:insert, Regex.match?(@insert_regex, query.statement)},
         {_, false} <- {:alter, Regex.match?(@alter_regex, query.statement)} do
      %{query | command: :unknown}
    else
      {command, true} -> %{query | command: command}
    end
  end

  defp parse_params_count(query) do
    params_count = @replace_regex |> Regex.scan(query.statement) |> Enum.count()
    %{query | params_count: params_count}
  end
end
